"""quixote.config

Configuration constants
"""

__revision__ = "$Id: config.py,v 1.13 2000/10/02 21:32:04 gward Exp $"

# Name of root package for the Web site

ROOT_PACKAGE = 'mems.ui'

# Prefix of URLs that will be directed to Quixote
URL_PREFIX = '/q'

# E-mail address to send errors to
# (None to send no mail at all)

ERROR_EMAIL = None
#ERROR_EMAIL = 'akuchlin@mems-exchange.org'

# Filename for logging debugging output (None for no log file.)

#DEBUG_LOG = None
DEBUG_LOG = "/tmp/quixote-debug.log"

# Page shown when there's an access-control violation
ACCESS_VIOLATION_PAGE = '/access_error'

# Filename for logging error messages (None for output to standard error,
# which should wind up in the Web servers log file.)

#ERROR_LOG = None
ERROR_LOG = "/www/httpd/logs/quixote.log"

# If true, exceptions will also be returned to the browser; otherwise,
# a generic error display will be used.

DISPLAY_EXCEPTIONS = 1

# If true, Quixote will service exactly one request at a time, and
# then exit.  This makes no difference when you're running as a
# straight CGI script, but it makes it easier to debug while running
# as a FastCGI script.
RUN_ONCE = 0

# If true, then any "resource not found" errors will result in a
# consistent, terse, mostly-useless message.  If false, then the
# exact cause of failure will be returned.
SECURE_ERRORS = 1

# Session-related variables
# =========================

# Domain to specify in setting cookies to hold the session ID
COOKIE_DOMAIN = '.cnri.reston.va.us'

# Name and path of the cookie that will hold the ID
COOKIE_NAME = "MX_session"
COOKIE_PATH = "/"
